@@ -1,15 +1,16 @@
 #!/usr/bin/env python3
+# -*- coding: utf-8 -*-
 #
-# Experimental implementation of asyncio support.
+# Implementation of asyncio support.
 #
 # This file is part of pySerial. https://github.com/pyserial/pyserial-asyncio
-# (C) 2015-2017 pySerial-team
+# (C) 2015-2020 pySerial-team
 #
 # SPDX-License-Identifier:    BSD-3-Clause
 """\
-Support asyncio with serial ports. EXPERIMENTAL
+Support asyncio with serial ports.
 
-Posix platforms only, Python 3.4+ only.
+Posix platforms only, Python 3.5+ only.
 
 Windows event loops can not wait for serial ports with the current
 implementation. It should be possible to get that working though.
@@ -24,7 +25,7 @@ try:
 except ImportError:
     termios = None
 
-__version__ = '0.4'
+__version__ = '0.5'
 
 
 class SerialTransport(asyncio.Transport):
@@ -74,9 +75,21 @@ class SerialTransport(asyncio.Transport):
 
     @property
     def serial(self):
-        """The underlying Serial instance."""
+        """The underlying Serial instance.
+
+        Equivalent to get_extra_info("serial")
+        """
         return self._serial
 
+    def get_extra_info(self, name, default=None):
+        """Get optional transport information.
+
+        Currently only "serial" is available.
+        """
+        if name == "serial":
+            return self._serial
+        return default
+
     def __repr__(self):
         return '{self.__class__.__name__}({self.loop}, {self._protocol}, {self.serial})'.format(self=self)
 
@@ -115,19 +128,11 @@ class SerialTransport(asyncio.Transport):
             return
 
         if self.get_write_buffer_size() == 0:
-            # Attempt to send it right away first
-            try:
-                n = self._serial.write(data)
-            except serial.SerialException as exc:
-                self._fatal_error(exc, 'Fatal write error on serial transport')
-                return
-            if n == len(data):
-                return  # Whole request satisfied
-            assert 0 <= n < len(data)
-            data = data[n:]
+            self._write_buffer.append(data)
             self._ensure_writer()
+        else:
+            self._write_buffer.append(data)
 
-        self._write_buffer.append(data)
         self._maybe_pause_protocol()
 
     def can_write_eof(self):
@@ -185,6 +190,13 @@ class SerialTransport(asyncio.Transport):
         """
         self._abort(None)
 
+    def flush(self):
+        """ clears output buffer and stops any more data being written
+        """
+        self._remove_writer()
+        self._write_buffer.clear()
+        self._maybe_resume_protocol()
+
     def _maybe_pause_protocol(self):
         """To be called whenever the write-buffer size increases.
 
@@ -272,9 +284,12 @@ class SerialTransport(asyncio.Transport):
     if os.name == "nt":
         def _poll_read(self):
             if self._has_reader:
-                if self.serial.in_waiting:
-                    self._loop.call_soon(self._read_ready)
-                self._loop.call_later(self._poll_wait_time, self._poll_read)
+                try:
+                    if self.serial.in_waiting:
+                        self._loop.call_soon(self._read_ready)
+                    self._loop.call_later(self._poll_wait_time, self._poll_read)
+                except serial.SerialException as exc:
+                    self._fatal_error(exc, 'Fatal write error on serial transport')
 
         def _ensure_reader(self):
             if (not self._has_reader) and (not self._closing):
@@ -386,15 +401,12 @@ class SerialTransport(asyncio.Transport):
         assert self._closing
         assert not self._has_writer
         assert not self._has_reader
-        if os.name == "nt":
+        try:
             self._serial.flush()
-        else:
-            try:
-                self._serial.flush()
-            except termios.error:
-                # ignore termios errors which may happen if the serial device was
-                # hot-unplugged.
-                pass
+        except (serial.SerialException if os.name == "nt" else termios.error):
+            # ignore serial errors which may happen if the serial device was
+            # hot-unplugged.
+            pass
         try:
             self._protocol.connection_lost(exc)
         finally:
@@ -405,18 +417,67 @@ class SerialTransport(asyncio.Transport):
             self._loop = None
 
 
-@asyncio.coroutine
-def create_serial_connection(loop, protocol_factory, *args, **kwargs):
-    ser = serial.serial_for_url(*args, **kwargs)
+async def create_serial_connection(loop, protocol_factory, *args, **kwargs):
+    """Create a connection to a new serial port instance.
+
+    This function is a coroutine which will try to establish the
+    connection.
+
+    The chronological order of the operation is:
+
+    1. protocol_factory is called without arguments and must return
+       a protocol instance.
+
+    2. The protocol instance is tied to the transport
+
+    3. This coroutine returns successfully with a (transport,
+       protocol) pair.
+
+    4. The connection_made() method of the protocol
+       will be called at some point by the event loop.
+
+    Note:  protocol_factory can be any kind of callable, not
+    necessarily a class. For example, if you want to use a pre-created
+    protocol instance, you can pass lambda: my_protocol.
+
+    Any additional arguments will be forwarded to the Serial constructor.
+    """
+    serial_instance = serial.serial_for_url(*args, **kwargs)
+    transport, protocol = await connection_for_serial(loop, protocol_factory, serial_instance)
+    return transport, protocol
+
+
+async def connection_for_serial(loop, protocol_factory, serial_instance):
+    """Create a connection to the given serial port instance.
+
+    This function is a coroutine which will try to establish the
+    connection.
+
+    The chronological order of the operation is:
+
+    1. protocol_factory is called without arguments and must return
+       a protocol instance.
+
+    2. The protocol instance is tied to the transport
+
+    3. This coroutine returns successfully with a (transport,
+       protocol) pair.
+
+    4. The connection_made() method of the protocol
+       will be called at some point by the event loop.
+
+    Note:  protocol_factory can be any kind of callable, not
+    necessarily a class. For example, if you want to use a pre-created
+    protocol instance, you can pass lambda: my_protocol.
+    """
     protocol = protocol_factory()
-    transport = SerialTransport(loop, protocol, ser)
-    return (transport, protocol)
+    transport = SerialTransport(loop, protocol, serial_instance)
+    return transport, protocol
 
 
-@asyncio.coroutine
-def open_serial_connection(*,
+async def open_serial_connection(*,
                            loop=None,
-                           limit=asyncio.streams._DEFAULT_LIMIT,
+                           limit=None,
                            **kwargs):
     """A wrapper for create_serial_connection() returning a (reader,
     writer) pair.
@@ -433,9 +494,11 @@ def open_serial_connection(*,
     """
     if loop is None:
         loop = asyncio.get_event_loop()
+    if limit is None:
+        limit = asyncio.streams._DEFAULT_LIMIT
     reader = asyncio.StreamReader(limit=limit, loop=loop)
     protocol = asyncio.StreamReaderProtocol(reader, loop=loop)
-    transport, _ = yield from create_serial_connection(
+    transport, _ = await create_serial_connection(
         loop=loop,
         protocol_factory=lambda: protocol,
         **kwargs)
@@ -477,6 +540,6 @@ if __name__ == '__main__':
 
     loop = asyncio.get_event_loop()
     coro = create_serial_connection(loop, Output, '/dev/ttyUSB0', baudrate=115200)
-    loop.run_until_complete(coro)
+    transport, protocol = loop.run_until_complete(coro)
     loop.run_forever()
     loop.close()
